﻿/*
* Copyright 2011 Benjamin Thomas Daniel Newman
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*       http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO.Ports;

namespace eyurale
{
    public partial class Form1 : Form
    {

        uint position_count = 0;
        bool locked = false; /* only handle one message on serial port at a time */
        SerialLog sl = new SerialLog();
        public Form1()
        {
            InitializeComponent();

            Text = "Euryale Desktop";

            PreconnectState();
            disconnectButton.Enabled = false;
            colorPanel.Hide();
            positionGroupBox.Hide();
            status("Ready to connect");
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ActiveForm.Close();
        }

        private void aboutEuryaleToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutBox1 about = new AboutBox1();
            about.ShowDialog();
        }

        private void connectButton_Click(object sender, EventArgs e)
        {
            int tmp;
            if (serialPort1.IsOpen)
            { /* should not get here */
                MessageBox.Show("Serial port is already open. Close it before trying to connect.", "", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }

            status("Connecting to CMUcam3...");

            if (comportBox.Text.Length == 0)
            {
                MessageBox.Show("COM Port not set.", "No COM Port", MessageBoxButtons.OK, MessageBoxIcon.Hand);
                status("No COM Port selected");
                return;
            }
            else
            {
                bool found = false;
                string[] serialports = SerialPort.GetPortNames();
                foreach (string i in serialports)
                {
                    if (i == comportBox.Text)
                        found = true;
                    break;
                }
                if (found)
                    serialPort1.PortName = comportBox.Text;
                else
                {
                    MessageBox.Show("COM Port " + comportBox.Text
                        + " is invalid. Available COM Ports:\n", "Bad COM Port", MessageBoxButtons.OK, MessageBoxIcon.Hand);
                    status("Invalid COM Port");
                    return;
                }
            }

            if (int.TryParse(baudBox.Text, out tmp))
                serialPort1.BaudRate = tmp;
            else
            {
                MessageBox.Show("Could not parse baud rate.", "Invalid Baud Rate", MessageBoxButtons.OK, MessageBoxIcon.Hand);
                status("Could not connect to CMUcam3: check baud rate");
                return;
            }

            serialPort1.DataBits = 8;
            serialPort1.Parity = Parity.None;
            serialPort1.StopBits = StopBits.One;
            serialPort1.Encoding = Encoding.Default;

            try
            {
                serialPort1.Open();
                serialPort1.DiscardInBuffer();
                serialPort1.DiscardOutBuffer();
            }
            catch
            {
                MessageBox.Show("Could not open serial port.", "Serial port error", MessageBoxButtons.OK, MessageBoxIcon.Hand);
                status("Could not open serial port");
                return;
            }

            status("Connected...");
            position_count = 0;
            connectButton.Enabled = false;
            disconnectButton.Enabled = true;
            baudBox.Enabled = false;
            comportBox.Enabled = false;
            timeoutBox.Enabled = false;
            sendDemoModeCommandToolStripMenuItem.Enabled = true;
            colorPanel.Show();
            positionGroupBox.Show();
            commandsToolStripMenuItem.Enabled = true;
            demoModeCheckBox.Enabled = true;
            demoModeCheckBox.Enabled = true;
            getPositionToolStripMenuItem.Enabled = true; /* has keyboard shortcut */
            trackNewColorToolStripMenuItem.Enabled = true;
        }

        private void disconnectButton_Click(object sender, EventArgs e)
        {
           if (serialPort1.IsOpen == true)
           {
               serialPort1.Close();
               serialPort1.Dispose();
           }

            status("Ready to reconnect");

            PreconnectState();
        }
        
        /* *_box, *_red, *_green and *_blue should be made into a class */
        private void min_color(int r, int g, int b)
        {

            if (this.InvokeRequired)
            {
                this.Invoke((MethodInvoker)delegate
                {
                    min_box.BackColor = Color.FromArgb(r, g, b);
                });
            }
            else
                min_box.BackColor = Color.FromArgb(r, g, b);
            SetLabel(min_red, r.ToString());
            SetLabel(min_green, g.ToString());
            SetLabel(min_blue, b.ToString());
        }

        private void max_color(int r, int g, int b)
        {
            if (this.InvokeRequired)
            {
                this.Invoke((MethodInvoker)delegate
                {
                    max_box.BackColor = Color.FromArgb(r, g, b);
                });
            }
            else
                max_box.BackColor = Color.FromArgb(r, g, b);
            SetLabel(max_red, r.ToString());
            SetLabel(max_green, g.ToString());
            SetLabel(max_blue, b.ToString());
        }

        private void pictureBox1_Click(object sender, EventArgs e)
        {
            MessageBox.Show("Cannot choose colour to track.\nOnly the CMUcam3 can choose the colour.", "Cannot choose colour", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        private void displaySerialPortSettingsToolStripMenuItem_CheckedChanged(object sender, EventArgs e)
        {
            if (displaySerialPortSettingsToolStripMenuItem.Checked)
                serialGroupBox.Show();
            else
                serialGroupBox.Hide();

        }

        void ProcessLine(string line)
        {
            sl.LogLine(line);
            sl.Show();
            if (EuryaleStrings.StartsWith(line, EuryaleStrings.LostObject()))
            {
                DefaultValues(false);
                status("Lost object!");
            }
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.Tracking()))
            {
                string centroid, boundingbox, numpix, density;

                centroid = EuryaleStrings.GetCentroid(line);
                boundingbox = EuryaleStrings.GetBoundingBox(line);
                numpix = EuryaleStrings.GetNumPix(line);
                density = EuryaleStrings.GetDensity(line);

                if (centroid.Length < 1)
                    MessageBox.Show("Malformed data from CMUcam3; missing centroid:\n"
                        + line, "Malformed data", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else if (boundingbox.Length < 1)
                    MessageBox.Show("Malformed data from CMUcam3; missing bounding box:\n"
                       + line, "Malformed data", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else if (numpix.Length < 1)
                    MessageBox.Show("Malformed data from CMUcam3; missing number of pixels:\n"
                       + line, "Malformed data", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else if (density.Length < 1)
                    MessageBox.Show("Malformed data from CMUcam3; missing density:\n"
                       + line, "Malformed data", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else
                {
                    position_count++;

                    SetLabelText(centroidLabel, centroid);

                    boundingboxLabel.Text = boundingbox;

                    UpdateAvgs(numpixelLabel, numpix, numpix_avgLabel, numpix_maxLabel, numpix_minLabel, position_count);
                    UpdateAvgs(densityLabel, density, density_avgLabel, density_maxLabel, density_minLabel, position_count);

                    status("Received new position data");
                }

            }
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.Color()))
            {
                string max, min;
                string[] tmp;
                char[] delim = { ',' };

                min = EuryaleStrings.GetColor(line, "min");
                max = EuryaleStrings.GetColor(line, "max");

                if (min.Length < 1)
                    MessageBox.Show("Malformed data from CMUcam3; minimum tracking color:\n"
                            + line, "Malformed data", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else if (max.Length < 1)
                    MessageBox.Show("Malformed data from CMUcam3; minimum tracking color:\n"
                            + line, "Malformed data", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else
                {
                    // Should do more error checking here

                    tmp = min.Split(delim);
                    SetLabel(min_red, tmp[0]);
                    SetLabel(min_green, tmp[1]);
                    SetLabel(min_green, tmp[2]);
                    min_color(int.Parse(min_red.Text),
                       int.Parse(min_green.Text),
                       int.Parse(min_green.Text));

                    tmp = max.Split(delim);
                    SetLabel(max_red, tmp[0]);
                    SetLabel(max_green, tmp[1]);
                    SetLabel(max_green, tmp[2]);
                    max_color(int.Parse(max_red.Text),
                       int.Parse(max_green.Text),
                       int.Parse(max_green.Text));
                }
            }
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.DemoMode()))
                demoModeCheckBox.Checked = false;
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.StartingUp()))
                status("Starting up...");
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.WaitingToStabalize()))
                status("Waiting for image to stabalize...");
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.RreadyForColor()))
                status("Press button to grab a color");
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.GrabbingColor()))
                status("Grabbing color...");
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.AskForDemo()))
                status("Press button to enter demo mode...");
            else if (EuryaleStrings.StartsWith(line, EuryaleStrings.InDemoMode()))
                status("Running in demo mode!");
            else
            {
                MessageBox.Show("Received unknown message:\n" + line, "Unknown message", MessageBoxButtons.OK, MessageBoxIcon.Information);
                status("Received unknown message");
            }
        }

        private void serialPort1_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            string line;
            SerialPort sp = (SerialPort)sender;

            if (locked)
                return;
            else
                locked = true;

            try
            {
                line = sp.ReadLine();
            }
            catch
            {
                MessageBox.Show("Could not read line from CMUcam3; try checking baud rate.", "Malformed data", MessageBoxButtons.OK, MessageBoxIcon.Hand);

                locked = false;
                return;
            }

            line.TrimEnd(Environment.NewLine.ToCharArray());

            if (line.Length < 1)
            {
                locked = false;
                return;
            }
 
                if (this.InvokeRequired)
                {
                    this.Invoke((MethodInvoker)delegate
                    {
                        ProcessLine(line);
                    });
                }
                else
                {
                    ProcessLine(line);
                }
   
            locked = false;
        }

        private void status(string Status)
        {
            if (this.InvokeRequired)
            {
                this.Invoke((MethodInvoker)delegate
                {
                    statusLabel.Text = Status;
                });
            }
            else
                statusLabel.Text = Status;

        }

        private void SetLabel(Label label, string Status)
        {
            if (this.InvokeRequired)
            {
                this.Invoke((MethodInvoker)delegate
                {
                    label.Text = Status;
                });
            }
            else
                label.Text = Status;
        }

        private void PopulateComPorts()
        {
            comportBox.Items.Clear();

            for (; ; )
            {
                foreach (string i in SerialPort.GetPortNames())
                    comportBox.Items.Add(i);

                if (comportBox.Items.Count == 0)
                {
                    DialogResult result = MessageBox.Show("No COM Ports are available.", "Euryale Desktop", MessageBoxButtons.AbortRetryIgnore, MessageBoxIcon.Stop);
                    if (result == DialogResult.Abort)
                        Environment.Exit(1);
                    else if (result == DialogResult.Retry)
                        continue;
                    else if (result == DialogResult.Ignore)
                        break;
                }
                else
                {
                    comportBox.SelectedIndex = 0;
                    break;
                }
            }
        }

        void PopulateBaudRates()
        {
            baudBox.Items.Clear();
            baudBox.Items.Add("115200");
            baudBox.Items.Add("57600");
            baudBox.Items.Add("38400");
            baudBox.Items.Add("28800");
            baudBox.Items.Add("14400");
            baudBox.Items.Add("9600");
            baudBox.Items.Add("4800");
            baudBox.Items.Add("2400");
            baudBox.Items.Add("1200");
            baudBox.SelectedIndex = 0;
        }

        void PreconnectState()
        {
            connectButton.Enabled = true;
            disconnectButton.Enabled = false;
            comportBox.Enabled = true;
            baudBox.Enabled = true;
            timeoutBox.Enabled = true;
            PopulateComPorts();
            PopulateBaudRates();
            demoModeCheckBox.Checked = false;
            sendDemoModeCommandToolStripMenuItem.Enabled = false;
            commandsToolStripMenuItem.Enabled = false;
            Reconnect();
            DisconnectState();
        }

        void DisconnectState()
        {
            connectButton.Enabled = true;
            disconnectButton.Enabled = false;
            comportBox.Enabled = true;
            baudBox.Enabled = true;
            timeoutBox.Enabled = true;

            demoModeCheckBox.Checked = false;
            demoModeCheckBox.Enabled = false;
        }

        void Reconnect()
        {
            min_color(0, 0, 0);
            max_color(255, 255, 255);
            DefaultValues(true);
        }


        private void DefaultValues(bool Reset)
        {
            centroidLabel.Text = "-1, -1";
            boundingboxLabel.Text = "-1, -1, -1, -1";
            numpixelLabel.Text = "-1";
            densityLabel.Text = "-1";

            if (Reset)
            {
                numpix_maxLabel.Text = "-1";
                numpix_minLabel.Text = "-1";
                density_maxLabel.Text = "-1";
                density_minLabel.Text = "-1";
                position_count = 0;
                numpix_avgLabel.Text = "-1";
                density_avgLabel.Text = "-1";
            }
        }

        /* C# 2010 introduces default parameters */
        void DefaultValues()
        {
            DefaultValues(true);
        }


        public void SetLabelText(Control control, string text)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new Action<Control, string>(SetLabelText), new object[] { control, text });
            }
            else
            {
                ((Control)control).Text = text;
            }
        }

        public void UpdateAvgs(Label cur, string val, Label avg, Label max, Label min, uint n)
        {
            SetLabel(cur, val);
            if (int.Parse(max.Text) < int.Parse(val))
                SetLabel(max, val);
            if (int.Parse(min.Text) == -1
                || int.Parse(min.Text) > int.Parse(val))
                SetLabel(min, val);
            if (n == 1)
                SetLabel(avg, val);
            else
                SetLabel(avg, ((uint.Parse(val) + ((n - 1) * uint.Parse(avg.Text)))
                    / n).ToString());
        }

        private void sendDemoModeCommandToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendMessage("AM");
        }

        private void getPositionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendMessage("PO");
        }

        private void sendMessage(string str){

            if (serialPort1.IsOpen)
                serialPort1.Write(str + "\r");
            else
                MessageBox.Show("Serial port is not open.");
        }
   
        private void getColorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendMessage("CC");
        }

        private void trackNewColorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendMessage("NC");
        }

        private void demoModeCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            if (((CheckBox)sender).Checked)
            {
                commandsToolStripMenuItem.Enabled = false;
                getPositionToolStripMenuItem.Enabled = false; /* has keyboard shortcut */
                trackNewColorToolStripMenuItem.Enabled = false;
            }
            else
            {
                commandsToolStripMenuItem.Enabled = true;
                getPositionToolStripMenuItem.Enabled = true;
                trackNewColorToolStripMenuItem.Enabled = true;
            }
        }

        private void openSerialLogToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sl.Show();
        }

        private void resetStatisticsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DefaultValues(true);
            position_count = 0;
        }

        private void sendDemoModeCommandToolStripMenuItem_Click_1(object sender, EventArgs e)
        {
            DialogResult result;

            result = MessageBox.Show("Are you sure you want to tell the CMUcam3 to enter automatic mode?"
                + "\n\nThe CMUcam3 will stop waiting for commands and will only give position data for the last color picked.",
                "Confirm entering automatic mode", MessageBoxButtons.YesNo, MessageBoxIcon.Question);

            if (result == DialogResult.No)
                return;

            sendMessage("AM");
            demoModeCheckBox.Checked = true;
        }

        private void resetToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendMessage("RS");
        }
    }
}